# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from enum import Enum
from typing import Literal

from redux import BaseAction, Immutable


class SoundDevice(str, Enum):
    INPUT = 'Input'
    OUTPUT = 'Output'


class SoundSetVolumeActionPayload(Immutable):
    volume: float
    device: SoundDevice


class SoundSetVolumeAction(BaseAction):
    type: Literal['SOUND_SET_VOLUME'] = 'SOUND_SET_VOLUME'
    payload: SoundSetVolumeActionPayload


class SoundChangeVolumeActionPayload(Immutable):
    amount: float
    device: SoundDevice


class SoundChangeVolumeAction(BaseAction):
    type: Literal['SOUND_CHANGE_VOLUME'] = 'SOUND_CHANGE_VOLUME'
    payload: SoundChangeVolumeActionPayload


class SoundSetMuteStatusActionPayload(Immutable):
    mute: bool
    device: SoundDevice


class SoundSetMuteStatusAction(BaseAction):
    type: Literal['SOUND_SET_MUTE_STATUS'] = 'SOUND_SET_MUTE_STATUS'
    payload: SoundSetMuteStatusActionPayload


class SoundToggleMuteStatusActionPayload(Immutable):
    device: SoundDevice


class SoundToggleMuteStatusAction(BaseAction):
    type: Literal['SOUND_TOGGLE_MUTE_STATUS'] = 'SOUND_TOGGLE_MUTE_STATUS'
    payload: SoundToggleMuteStatusActionPayload


SoundAction = (
    SoundSetVolumeAction
    | SoundChangeVolumeAction
    | SoundSetMuteStatusAction
    | SoundToggleMuteStatusAction
)


class SoundState(Immutable):
    output_volume: float
    is_output_mute: bool
    mic_volume: float
    is_mic_mute: bool



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.sound import (
    SoundAction,
    SoundDevice,
    SoundSetMuteStatusAction,
    SoundSetMuteStatusActionPayload,
    SoundState,
)
from ubo_app.store.status_icons import (
    IconRegistrationAction,
    IconRegistrationActionPayload,
)

Action = InitAction | SoundAction | IconRegistrationAction


def reducer(
    state: SoundState | None,
    action: Action,
) -> ReducerResult[SoundState, Action]:
    if state is None:
        if action.type == 'INIT':
            return SoundState(
                output_volume=0.5,
                is_output_mute=False,
                mic_volume=0.5,
                is_mic_mute=False,
            )
        raise InitializationActionError

    if action.type == 'SOUND_SET_VOLUME':
        if action.payload.device == SoundDevice.OUTPUT:
            return replace(state, output_volume=action.payload.volume)
        if action.payload.device == SoundDevice.INPUT:
            return replace(state, input_volume=action.payload.volume)
    elif action.type == 'SOUND_CHANGE_VOLUME':
        if action.payload.device == SoundDevice.OUTPUT:
            return replace(
                state,
                output_volume=min(
                    max(state.output_volume + action.payload.amount, 0),
                    1,
                ),
            )
        if action.payload.device == SoundDevice.INPUT:
            return replace(
                state,
                input_volume=min(
                    max(state.mic_volume + action.payload.amount, 0),
                    1,
                ),
            )
    elif action.type == 'SOUND_SET_MUTE_STATUS':
        if action.payload.device == SoundDevice.OUTPUT:
            return replace(state, is_output_mute=action.payload.mute)
        if action.payload.device == SoundDevice.INPUT:
            return CompleteReducerResult(
                state=replace(state, is_mic_mute=action.payload.mute),
                actions=[
                    IconRegistrationAction(
                        payload=IconRegistrationActionPayload(
                            icon='mic_off' if action.payload.mute else 'mic',
                            priority=-2,
                            id='sound_mic_status',
                        ),
                    ),
                ],
            )
    elif action.type == 'SOUND_TOGGLE_MUTE_STATUS':
        return CompleteReducerResult(
            state=state,
            actions=[
                SoundSetMuteStatusAction(
                    payload=SoundSetMuteStatusActionPayload(
                        mute=not state.is_output_mute
                        if action.payload.device == SoundDevice.OUTPUT
                        else not state.is_mic_mute,
                        device=action.payload.device,
                    ),
                ),
            ],
        )
    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107

from ubo_app.load_services import load_services
from ubo_app.menu import MenuApp


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    app = MenuApp()

    load_services()

    app.run()


if __name__ == '__main__':
    main()



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import field, replace
from typing import TYPE_CHECKING, Literal

from redux import (
    BaseAction,
    CompleteReducerResult,
    Immutable,
    InitAction,
    InitializationActionError,
    ReducerResult,
)
from ubo_gui.menu import Item, is_sub_menu_item, menu_items

from ubo_app.store.app import RegisterAppAction, is_app_registration_action
from ubo_app.store.keypad import (
    Key,
    KeyEvent,
    KeyEventPayload,
    KeypadAction,
    KeyPressEvent,
)
from ubo_app.store.sound import (
    SoundChangeVolumeAction,
    SoundChangeVolumeActionPayload,
    SoundDevice,
)

from ._menus import HOME_MENU

if TYPE_CHECKING:
    from typing_extensions import TypeAlias
    from ubo_gui.menu.types import Menu
    from ubo_gui.page import PageWidget


class MainState(Immutable):
    current_menu: Menu | None = None
    current_application: type[PageWidget] | None = None
    path: list[int] = field(default_factory=list)


class SetMenuPathActionPayload(Immutable):
    path: list[str]


class SetMenuPathAction(BaseAction):
    type: Literal['MAIN_SET_MENU_PATH'] = 'MAIN_SET_MENU_PATH'
    payload: SetMenuPathActionPayload


MainAction: TypeAlias = (
    InitAction | KeypadAction | RegisterAppAction | SetMenuPathAction
)


def main_reducer(
    state: MainState | None,
    action: MainAction,
) -> ReducerResult[MainState, SoundChangeVolumeAction, KeyEvent]:
    if state is None:
        if action.type == 'INIT':
            return MainState(current_menu=HOME_MENU)
        raise InitializationActionError

    if action.type == 'KEYPAD_KEY_PRESS':
        actions = []
        if action.payload.key == Key.UP and len(state.path) == 0:
            actions.append(
                SoundChangeVolumeAction(
                    payload=SoundChangeVolumeActionPayload(
                        amount=0.05,
                        device=SoundDevice.OUTPUT,
                    ),
                ),
            )
        if action.payload.key == Key.DOWN and len(state.path) == 0:
            actions.append(
                SoundChangeVolumeAction(
                    payload=SoundChangeVolumeActionPayload(
                        amount=-0.05,
                        device=SoundDevice.OUTPUT,
                    ),
                ),
            )
        return CompleteReducerResult(
            state=state,
            actions=actions,
            events=[KeyPressEvent(payload=KeyEventPayload(key=action.payload.key))],
        )

    if is_app_registration_action(action):
        # TODO(sassanh): clone the menu
        # menu = copy.deepcopy(state.current_menu)
        menu = state.current_menu

        main_menu_item: Item = menu_items(menu)[0]
        if not is_sub_menu_item(main_menu_item):
            msg = 'Main menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        container_menu_item: Item
        if action.type == 'MAIN_REGISTER_REGULAR_APP':
            container_menu_item = menu_items(main_menu_item['sub_menu'])[0]
        else:
            container_menu_item = menu_items(main_menu_item['sub_menu'])[1]

        if not is_sub_menu_item(container_menu_item):
            msg = 'Settings menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        menu_items(container_menu_item['sub_menu']).append(action.payload.menu_item)
        return replace(state, current_menu=menu)

    if action.type == 'MAIN_SET_MENU_PATH':
        return replace(state, path=action.payload.path)

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import os

from headless_kivy_pi import setup_headless

os.environ['KIVY_METRICS_DENSITY'] = '1'
os.environ['KIVY_NO_CONFIG'] = '1'
os.environ['KIVY_NO_FILELOG'] = '1'

setup_headless()

from ubo_gui.app import UboApp  # noqa: E402

from .menu_central import MenuAppCentral  # noqa: E402
from .menu_footer import MenuAppFooter  # noqa: E402


class MenuApp(MenuAppCentral, MenuAppFooter, UboApp):
    """Menu application."""



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from enum import Enum
from typing import Literal

from redux import BaseAction, BaseEvent, Immutable


class Key(str, Enum):
    BACK = 'Back'
    HOME = 'Home'
    UP = 'Up'
    DOWN = 'Down'
    L1 = 'L1'
    L2 = 'L2'
    L3 = 'L3'


class KeypadActionPayload(Immutable):
    key: Key


KeyType = Literal['KEYPAD_KEY_UP', 'KEYPAD_KEY_DOWN', 'KEYPAD_KEY_PRESS']


class KeypadAction(BaseAction):
    payload: KeypadActionPayload
    type: KeyType


class KeypadKeyUpAction(KeypadAction):
    type: Literal['KEYPAD_KEY_UP'] = 'KEYPAD_KEY_UP'


class KeypadKeyDownAction(KeypadAction):
    type: Literal['KEYPAD_KEY_DOWN'] = 'KEYPAD_KEY_DOWN'


class KeypadKeyPressAction(KeypadAction):
    type: Literal['KEYPAD_KEY_PRESS'] = 'KEYPAD_KEY_PRESS'


class KeyEventPayload(Immutable):
    key: Key


class BaseKeyEvent(BaseEvent):
    payload: KeyEventPayload
    type: KeyType


class KeyPressEvent(BaseKeyEvent):
    type: Literal['KEYPAD_KEY_PRESS'] = 'KEYPAD_KEY_PRESS'


KeyEvent = KeyPressEvent


def dispatch_key(action: KeyType, key: Key) -> None:
    from ubo_app.store import dispatch

    cls: type[KeypadAction]

    if action == 'KEYPAD_KEY_DOWN':
        cls = KeypadKeyDownAction
    elif action == 'KEYPAD_KEY_UP':
        cls = KeypadKeyUpAction
    elif action == 'KEYPAD_KEY_PRESS':
        cls = KeypadKeyPressAction

    dispatch(cls(payload=KeypadActionPayload(key=key)))



# ruff: noqa: D100, D101, D102, D103, D104, D107
from typing import Literal, TypeGuard

from redux import BaseAction, Immutable
from ubo_gui.menu import Item


class RegisterAppActionPayload(Immutable):
    menu_item: Item


RegisterAppActionType = Literal[
    'MAIN_REGISTER_REGULAR_APP',
    'MAIN_REGISTER_SETTING_APP',
]


class RegisterAppAction(BaseAction):
    payload: RegisterAppActionPayload
    type: RegisterAppActionType


class RegisterRegularAppAction(RegisterAppAction):
    type: Literal['MAIN_REGISTER_REGULAR_APP'] = 'MAIN_REGISTER_REGULAR_APP'


class RegisterSettingAppAction(RegisterAppAction):
    type: Literal['MAIN_REGISTER_SETTING_APP'] = 'MAIN_REGISTER_SETTING_APP'


def is_app_registration_action(action: BaseAction) -> TypeGuard[RegisterAppAction]:
    return action.type in ('MAIN_REGISTER_REGULAR_APP', 'MAIN_REGISTER_SETTING_APP')



from ubo_gui.menu.types import Menu
from ubo_gui.notification import notification_manager

SETTINGS_MENU: Menu = {
    'title': 'Settings',
    'items': [],
}

APPS_MENU: Menu = {
    'title': 'Apps',
    'items': [],
}

MAIN_MENU: Menu = {
    'title': 'Main',
    'items': [
        {
            'label': 'Apps',
            'icon': 'apps',
            'sub_menu': APPS_MENU,
        },
        {
            'label': 'Settings',
            'icon': 'settings',
            'sub_menu': SETTINGS_MENU,
        },
        {
            'label': 'About',
            'action': lambda: print('About'),
            'icon': 'info',
        },
    ],
}

HOME_MENU: Menu = {
    'title': 'Dashboard',
    'items': [
        {
            'label': '',
            'sub_menu': MAIN_MENU,
            'icon': 'menu',
            'is_short': True,
        },
        {
            'label': '',
            'sub_menu': {
                'title': lambda: f'Notifications ({notification_manager.unread_count})',
                'items': notification_manager.menu_items,
            },
            'color': 'yellow',
            'icon': 'info',
            'is_short': True,
        },
        {
            'label': 'Turn off',
            'action': lambda: print('Turning off'),
            'icon': 'power_settings_new',
            'is_short': True,
        },
    ],
}



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import TYPE_CHECKING

from redux import (
    BaseCombineReducerState,
    CombineReducerAction,
    InitAction,
    combine_reducers,
    create_store,
)

from ubo_app.store.keypad import KeyEvent
from ubo_app.store.main import MainAction, MainState, main_reducer
from ubo_app.store.sound import SoundAction, SoundState

from .status_icons import IconAction
from .status_icons import reducer as status_icons_reducer

if TYPE_CHECKING:
    from .status_icons import StatusIconsState


class RootState(BaseCombineReducerState):
    main: MainState
    sound: SoundState
    status_icons: StatusIconsState


root_reducer, root_reducer_id = combine_reducers(
    state_type=RootState,
    action_type=MainAction | SoundAction | IconAction | CombineReducerAction,
    event_type=KeyEvent,
    main=main_reducer,
    status_icons=status_icons_reducer,
)


store = create_store(root_reducer)

store.dispatch(InitAction(type='INIT'))

autorun = store.autorun


def dispatch(items: ActionType | EventType | list[ActionType | EventType]) -> None:
    from kivy.clock import Clock

    logger.debug(items)
    Clock.schedule_once(lambda _: store.dispatch(items), -1)


subscribe = store.subscribe
subscribe_event = store.subscribe_event

__ALL__ = (autorun, dispatch, subscribe)



# ruff: noqa: D100, D101, D102, D103, D104, D107
import importlib
import importlib.util
import os
import sys
from pathlib import Path
from typing import Any

from redux import CombineReducerRegisterAction, CombineReducerRegisterActionPayload

from ubo_app.logging import logger

ROOT_PATH = Path(__file__).parent


def load(path: Path) -> Any:
    try:
        if path.exists():
            spec = importlib.util.spec_from_file_location(
                '__ubo_service__',
                location=path.as_posix(),
                submodule_search_locations=[path.parent.as_posix()],
            )
            if not spec:
                return None
            module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = module
            if spec.loader:
                spec.loader.exec_module(module)
            return module
    except Exception as exception:  # noqa: BLE001
        logger.error(f'Error loading "{path}"')
        logger.exception(exception)


def load_services() -> None:
    from ubo_app import store

    for services_directory_path in [
        ROOT_PATH.joinpath('services').as_posix(),
        *(
            os.environ.get('UBO_SERVICES_PATH', '').split(':')
            if os.environ.get('UBO_SERVICES_PATH')
            else []
        ),
    ]:
        if Path(services_directory_path).is_dir():
            for service_path in Path(services_directory_path).iterdir():
                if not service_path.is_dir():
                    continue
                current_path = os.curdir
                os.chdir(service_path.as_posix())

                info = load(service_path.joinpath('__init__.py'))
                reducer = load(service_path.joinpath('reducer.py'))
                load(service_path.joinpath('setup.py'))

                if reducer is not None and hasattr(reducer, 'reducer'):
                    store.dispatch(
                        CombineReducerRegisterAction(
                            _id=store.root_reducer_id,
                            payload=CombineReducerRegisterActionPayload(
                                key=info.ubo_service_id,
                                reducer=reducer.reducer,
                            ),
                        ),
                    )

                os.chdir(current_path)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from ubo_gui.menu.types import ApplicationItem
from ubo_gui.notification import Importance, notification_manager
from ubo_gui.page import PageWidget
from ubo_gui.prompt import PromptWidget

from ubo_app.store import dispatch
from ubo_app.store.app import RegisterAppActionPayload, RegisterSettingAppAction
from ubo_app.store.status_icons import (
    IconRegistrationAction,
    IconRegistrationActionPayload,
)

ubo_service_name = 'WiFi'
ubo_service_description = 'WiFi app for ubo-pod'


class WifiSetup(PageWidget):
    pass


class WifiPrompt(PromptWidget):
    icon = 'wifi_off'
    prompt = 'Not Connected'
    first_option_label = 'Add'
    first_option_icon = 'add'

    def first_option_callback(self: WifiPrompt) -> type[PageWidget]:
        return WifiSetup

    second_option_label = 'Forget'
    second_option_icon = 'delete'

    def second_option_callback(self: WifiPrompt) -> None:
        notification_manager.notify(
            title='WiFi forgotten',
            content='This WiFi network is forgotten',
            importance=Importance.CRITICAL,
            sender='WiFi',
        )


def init_service() -> None:
    dispatch(
        [
            RegisterSettingAppAction(
                payload=RegisterAppActionPayload(
                    menu_item=ApplicationItem(
                        label='WiFi',
                        application=WifiPrompt,
                        icon='wifi',
                    ),
                ),
            ),
            RegisterSettingAppAction(
                payload=RegisterAppActionPayload(
                    menu_item=ApplicationItem(
                        label='WiFi',
                        application=WifiPrompt,
                        icon='wifi',
                    ),
                ),
            ),
            RegisterSettingAppAction(
                payload=RegisterAppActionPayload(
                    menu_item=ApplicationItem(
                        label='WiFi',
                        application=WifiPrompt,
                        icon='wifi',
                    ),
                ),
            ),
            RegisterSettingAppAction(
                payload=RegisterAppActionPayload(
                    menu_item=ApplicationItem(
                        label='WiFi',
                        application=WifiPrompt,
                        icon='wifi',
                    ),
                ),
            ),
            RegisterSettingAppAction(
                payload=RegisterAppActionPayload(
                    menu_item=ApplicationItem(
                        label='WiFi',
                        application=WifiPrompt,
                        icon='wifi',
                    ),
                ),
            ),
            IconRegistrationAction(
                payload=IconRegistrationActionPayload(icon='wifi', priority=-1),
            ),
        ],
    )


if __name__ == '__ubo_service__':
    init_service()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from ubo_app.store import dispatch
from ubo_app.store.status_icons import (
    IconRegistrationAction,
    IconRegistrationActionPayload,
)

ubo_service_id = 'sound'
ubo_service_name = 'Sound'
ubo_service_description = 'Sound control for ubo-pod'


def init_service() -> None:
    dispatch(
        IconRegistrationAction(
            payload=IconRegistrationActionPayload(
                icon='mic_off',
                priority=-2,
                id='sound_mic_status',
            ),
        ),
    )


if __name__ == '__ubo_service__':
    init_service()
